/*  Gpasman, a password manager
    Copyright (C) 1998, 1999 Olivier Sessink, olivier@lx.student.wau.nl

    Other code contributors:    
    Dave Rudder 
    Chris Halverson
    Matthew Palmer
    Guide Berning
    Jimmy Mason
	 website at http://www.student.wau.nl/~olivier/gpasman/

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <gtk/gtk.h>
#include <glib.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include "config.h"
#include "rc2.h"

#ifndef S_IAMB
#define S_IAMB 00777
#endif

typedef struct _gpasman_set
  {
    gchar *host;
    gchar *username;
    gchar *password;
    gchar *comment;
  }
Gpasman_Set;

void button_add_clicked (GtkWidget * button, gpointer data);
void button_change_clicked (GtkWidget * button, gpointer data);
void button_clear_clicked (GtkWidget * button, gpointer data);
void really_clear_list (GtkWidget * button, gpointer data);
void button_hide_show_clicked (GtkWidget * button, gpointer data);
void selection_made (GtkWidget * clist, gint row, gint column,
		     GdkEventButton * event, gpointer data);
void unselection_made (GtkWidget * clist, gint row, gint column,
		       GdkEventButton * event, gpointer data);
void click_made (GtkWidget * clist, gint column, gpointer data);
void new_set_add (GtkWidget * button, gpointer data);
void change_set (GtkWidget * button, gpointer data);
void window_close (GtkWidget * button, gpointer windowname);
void quit_gpasman (GtkObject * object);
void button_remove_clicked (GtkWidget * button, gpointer data);
void save_all (GtkWidget * button, gpointer end);
int openfile ();
void main_window (GtkWidget * oldwindow, gpointer data);
void change_password (GtkWidget * button, gpointer data);
void check_password (GtkWidget * button, gpointer data);
void warning_dialog (char *warningtext, char quit);
int check_gpasfile (char *passfile);
int file_exists (char *tfile);
int compare_set (const void *, const void *);
int create_gpasfile ();
int create_swpfile ();
int remove_swpfile ();
int is_running ();

GtkWidget *clist, *entry1, *entry2, *entry3, *entry4;
int selectedrow = -1, totalrows = 0;
char password[30];
char naam[256];
char swp_file[256];
char changed = 0;
char end = 1, noend = 0;
int sort_key = 0;

gint
main (int argc, gchar * argv[])
{
  GtkWidget *window;
  GtkWidget *vbox;
  GtkWidget *button_ok;
  GtkWidget *label1;
  int tmpInt;

  gtk_init (&argc, &argv);

  tmpInt = strlen (getenv ("HOME"));
  strncpy (naam, getenv ("HOME"), tmpInt);
  tmpInt = strlen ("/.gpasman");
  strncat (naam, "/.gpasman", tmpInt);

  if (file_exists (naam))
    {
      warning_dialog ("~/.gpasman does not exist, will create a new one", 0);
      create_gpasfile ();
    }

  tmpInt = strlen (getenv ("HOME"));
  strncpy (swp_file, getenv ("HOME"), tmpInt);
  tmpInt = strlen ("/.gpasman.swp");
  strncat (swp_file, "/.gpasman.swp", tmpInt);

  if (is_running () == 0)
    {
/*      atexit ( remove_swpfile() ); */
      create_swpfile ();
    }
  else
    {
        warning_dialog ("Gpassman is already running (~/.gpasman.swp found). \n Quit other gpasman or remove  ~/.gpasman.swp file.", 1);
        gtk_main ();
        return (0);
    }

  if (check_gpasfile (naam))
    {
      warning_dialog
		      ("There is an error with your ~/.gpasman file, look at the stderr output", 1);
      gtk_main ();
    }
  else
    {
      window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
      gtk_container_border_width (GTK_CONTAINER (window), 10);
      gtk_widget_set_usize (GTK_WIDGET (window), 200, 150);
      gtk_window_set_title (GTK_WINDOW (window), "Gpasman 1.1.3");
      vbox = gtk_vbox_new (FALSE, 5);
      gtk_container_border_width (GTK_CONTAINER (vbox), 5);
      gtk_container_add (GTK_CONTAINER (window), vbox);
      gtk_widget_show (vbox);

      entry1 = gtk_entry_new_with_max_length (100);
      gtk_entry_set_visibility (GTK_ENTRY (entry1), 0);
      button_ok = gtk_button_new_with_label ("Ok");
      label1 = gtk_label_new ("Enter password");

      gtk_box_pack_start (GTK_BOX (vbox), label1, TRUE, TRUE, 3);
      gtk_box_pack_start (GTK_BOX (vbox), entry1, TRUE, TRUE, 3);
      gtk_box_pack_start (GTK_BOX (vbox), button_ok, TRUE, TRUE, 3);

      gtk_signal_connect (GTK_OBJECT (window),
			  "destroy",
			  GTK_SIGNAL_FUNC (gtk_main_quit),
			  NULL);
      /* we have to remove the lockfile too!! */
			  
      gtk_signal_connect_object (GTK_OBJECT (entry1), "activate",
				 GTK_SIGNAL_FUNC (gtk_widget_grab_focus),
				 (gpointer) button_ok);
      gtk_widget_show (button_ok);
      gtk_widget_show (label1);
      gtk_widget_show (entry1);
      gtk_widget_show (window);
      gtk_signal_connect (GTK_OBJECT (button_ok), "clicked",
			  GTK_SIGNAL_FUNC (main_window),
			  window);
      gtk_widget_grab_focus (entry1);
      gtk_main ();
    }
  return 0;
}

void
main_window (GtkWidget * widget, gpointer oldwindow)
{
  GtkWidget *window, *vbox, *hbox1, *hbox2, *scrolwin;
  GtkWidget *button_add, *button_change, *button_clear, *button_remove,
  *button_save, *button_exit, *button_chp;
  gchar *titles[4] =
  {"Host", "Username", "Password", "Comment"};
  int tmpInt;

  /* first make the entry1 ready for re-use */

  tmpInt = strlen (gtk_entry_get_text (GTK_ENTRY (entry1)));

  strncpy (password, gtk_entry_get_text (GTK_ENTRY (entry1)), tmpInt);

  gtk_signal_handlers_destroy ((GtkObject *) oldwindow);
  gtk_widget_destroy (oldwindow);
  window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
  gtk_widget_set_usize (GTK_WIDGET (window), 600, 300);

  gtk_window_set_title (GTK_WINDOW (window), "Gpasman 1.1.3");
  gtk_signal_connect (GTK_OBJECT (window),
		      "destroy",
		      GTK_SIGNAL_FUNC (gtk_main_quit),
		      NULL);

      /* we have to remove the lockfile too!! */

  vbox = gtk_vbox_new (FALSE, 5);
  gtk_container_border_width (GTK_CONTAINER (vbox), 5);
  gtk_container_add (GTK_CONTAINER (window), vbox);
  gtk_widget_show (vbox);

  /* since GTK 1.2 one should use a scrolledwindow for a clist */
  scrolwin = gtk_scrolled_window_new (NULL, NULL);
  gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (scrolwin),
				  GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
  gtk_box_pack_start (GTK_BOX (vbox), scrolwin, TRUE, TRUE, 0);
  gtk_widget_show (scrolwin);
  /* Create the GtkCList */
  clist = gtk_clist_new_with_titles (4, titles);
  gtk_clist_column_titles_active (GTK_CLIST (clist));
  gtk_signal_connect (GTK_OBJECT (clist), "select_row",
		      GTK_SIGNAL_FUNC (selection_made),
		      NULL);
  gtk_signal_connect (GTK_OBJECT (clist), "unselect_row",
		      GTK_SIGNAL_FUNC (unselection_made),
		      NULL);
  gtk_signal_connect (GTK_OBJECT (clist), "click_column",
		      GTK_SIGNAL_FUNC (click_made),
		      NULL);

  /* It isn't necessary to shadow the border, but it looks nice :) */
  gtk_clist_set_shadow_type (GTK_CLIST (clist), GTK_SHADOW_OUT);
  gtk_clist_set_column_width (GTK_CLIST (clist), 0, 200);
  gtk_clist_set_column_width (GTK_CLIST (clist), 1, 100);
  gtk_clist_set_column_width (GTK_CLIST (clist), 2, 100);
  gtk_clist_set_column_width (GTK_CLIST (clist), 3, 200);

  /* Scollbars _only when needed_ */
/*	gtk_clist_set_policy(GTK_CLIST(clist), GTK_POLICY_AUTOMATIC,
                                                GTK_POLICY_AUTOMATIC); */

  gtk_scrolled_window_add_with_viewport (GTK_SCROLLED_WINDOW (scrolwin), clist);
  gtk_widget_show (clist);

  hbox1 = gtk_hbox_new (TRUE, 0);
  hbox2 = gtk_hbox_new (TRUE, 0);
  gtk_box_pack_start (GTK_BOX (vbox), hbox1, FALSE, TRUE, 0);
  gtk_box_pack_start (GTK_BOX (vbox), hbox2, FALSE, TRUE, 0);
  gtk_widget_show (hbox1);
  gtk_widget_show (hbox2);

  button_add = gtk_button_new_with_label ("Add");
  button_change = gtk_button_new_with_label ("Change");
  button_clear = gtk_button_new_with_label ("Clear List");
  button_remove = gtk_button_new_with_label ("Remove");
  button_save = gtk_button_new_with_label ("Save");
  button_exit = gtk_button_new_with_label ("Exit");
  button_chp = gtk_button_new_with_label ("Change password");
  gtk_box_pack_start (GTK_BOX (hbox1), button_add, TRUE, TRUE, 0);
  gtk_box_pack_start (GTK_BOX (hbox1), button_change, TRUE, TRUE, 0);
  gtk_box_pack_start (GTK_BOX (hbox1), button_remove, TRUE, TRUE, 0);
  gtk_box_pack_start (GTK_BOX (hbox1), button_clear, TRUE, TRUE, 0);
  gtk_box_pack_start (GTK_BOX (hbox2), button_chp, TRUE, TRUE, 0);
  gtk_box_pack_start (GTK_BOX (hbox2), button_save, TRUE, TRUE, 0);
  gtk_box_pack_start (GTK_BOX (hbox2), button_exit, TRUE, TRUE, 0);

  gtk_signal_connect_object (GTK_OBJECT (button_add), "clicked",
			     GTK_SIGNAL_FUNC (button_add_clicked),
			     (gpointer) clist);
  gtk_signal_connect_object (GTK_OBJECT (button_change), "clicked",
			     GTK_SIGNAL_FUNC (button_change_clicked),
			     (gpointer) clist);
  gtk_signal_connect (GTK_OBJECT (button_clear), "clicked",
		      GTK_SIGNAL_FUNC (button_clear_clicked),
		      (gpointer) clist);
  gtk_signal_connect_object (GTK_OBJECT (button_remove), "clicked",
			     GTK_SIGNAL_FUNC (button_remove_clicked),
			     (gpointer) clist);
  gtk_signal_connect (GTK_OBJECT (button_save), "clicked",
		      GTK_SIGNAL_FUNC (save_all),
		      (gpointer) & noend);
  gtk_signal_connect_object (GTK_OBJECT (button_exit), "clicked",
			     GTK_SIGNAL_FUNC (quit_gpasman),
			     (gpointer) window);
  gtk_signal_connect (GTK_OBJECT (button_chp), "clicked",
		      GTK_SIGNAL_FUNC (change_password),
		      NULL);

  gtk_widget_show (button_add);
  gtk_widget_show (button_change);
  gtk_widget_show (button_clear);
  gtk_widget_show (button_remove);
  gtk_widget_show (button_save);
  gtk_widget_show (button_exit);
  gtk_widget_show (button_chp);
  gtk_widget_show (window);
  openfile ();
}

void
quit_gpasman (GtkObject * object)
{
  GtkWidget *window, *label1, *button_ok, *button_cancel, *vbox, *hbox, *button_quit;

  if (changed)
    {
      window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
      gtk_container_border_width (GTK_CONTAINER (window), 10);
      gtk_grab_add (window);
      gtk_window_position (GTK_WINDOW (window), GTK_WIN_POS_MOUSE);
      gtk_window_set_title (GTK_WINDOW (window), "Really exit?");
      gtk_signal_connect (GTK_OBJECT (window), "delete_event",
			  GTK_SIGNAL_FUNC (gtk_main_quit), NULL);
      vbox = gtk_vbox_new (TRUE, 0);
      gtk_container_add (GTK_CONTAINER (window), vbox);
      gtk_widget_show (vbox);
      label1 = gtk_label_new ("There is unsaved data, save?");
      gtk_box_pack_start (GTK_BOX (vbox), label1, TRUE, TRUE, 0);
      gtk_widget_show (label1);

      hbox = gtk_hbox_new (TRUE, 0);
      gtk_container_add (GTK_CONTAINER (vbox), hbox);
      gtk_widget_show (hbox);

      button_ok = gtk_button_new_with_label ("Save");
      button_cancel = gtk_button_new_with_label ("Cancel");
      button_quit = gtk_button_new_with_label ("Quit");
      gtk_box_pack_start (GTK_BOX (hbox), button_ok, TRUE, TRUE, 0);
      gtk_widget_show (button_ok);
      gtk_box_pack_start (GTK_BOX (hbox), button_quit, TRUE, TRUE, 0);
      gtk_widget_show (button_quit);
      gtk_box_pack_start (GTK_BOX (hbox), button_cancel, TRUE, TRUE, 0);
      gtk_widget_show (button_cancel);
      gtk_signal_connect (GTK_OBJECT (button_quit), "clicked",
			  GTK_SIGNAL_FUNC (gtk_main_quit),
			  NULL);
      gtk_signal_connect (GTK_OBJECT (button_cancel), "clicked",
			  GTK_SIGNAL_FUNC (window_close),
			  window);
      gtk_signal_connect_object (GTK_OBJECT (button_ok), "clicked",
				 GTK_SIGNAL_FUNC (save_all),
				 (gpointer) & end);
      gtk_widget_show (window);
    }
  else
    {
      gtk_main_quit ();
    }
  remove_swpfile ();

}

void
new_set_add (GtkWidget * button, gpointer data)
{

  gchar *entry_text[4];
  gint count;

  entry_text[0] = gtk_entry_get_text (GTK_ENTRY (entry1));
  entry_text[1] = gtk_entry_get_text (GTK_ENTRY (entry2));
  entry_text[2] = gtk_entry_get_text (GTK_ENTRY (entry3));
  entry_text[3] = gtk_entry_get_text (GTK_ENTRY (entry4));
  for (count =0 ; count < 4 ; count++) {
#ifdef DEBUG
     g_print("new_set_add, count=%d\n", count);
#endif
     if (strlen(entry_text[count]) == 0) {
        strcpy(entry_text[count], " ");
#ifdef DEBUG
        g_print("new_set_add, strlen(entry_text[%d])==0\n",count);
#endif
     }
  }
  gtk_clist_append ((GtkCList *) clist, entry_text);
  totalrows++;
  changed = 1;
  window_close (button, data);
}

void
change_set (GtkWidget * button, gpointer data)
{

  gchar *entry_text[4];
  gint count;

  entry_text[0] = gtk_entry_get_text (GTK_ENTRY (entry1));
  entry_text[1] = gtk_entry_get_text (GTK_ENTRY (entry2));
  entry_text[2] = gtk_entry_get_text (GTK_ENTRY (entry3));
  entry_text[3] = gtk_entry_get_text (GTK_ENTRY (entry4));
  for (count =0 ; count < 4 ; count++) {
#ifdef DEBUG
     g_print("change_set, count=%d\n", count);
#endif
     if (strlen(entry_text[count]) == 0) {
        strcpy(entry_text[count], " ");
#ifdef DEBUG
        g_print("change_set, strlen(entry_text[%d])==0\n",count);
#endif
     }
  }
  gtk_clist_set_text (GTK_CLIST (clist), selectedrow, 0, entry_text[0]);
  gtk_clist_set_text (GTK_CLIST (clist), selectedrow, 1, entry_text[1]);
  gtk_clist_set_text (GTK_CLIST (clist), selectedrow, 2, entry_text[2]);
  gtk_clist_set_text (GTK_CLIST (clist), selectedrow, 3, entry_text[3]);
  changed = 1;
  window_close (button, data);
}

void
window_close (GtkWidget * button, gpointer windowname)
{
#ifdef DEBUG
  g_print ("window_close, windowname=%p\n", windowname);
#endif
  gtk_grab_remove (windowname);
  gtk_widget_destroy (windowname);
}

void
button_remove_clicked (GtkWidget * button, gpointer data)
{
  if (selectedrow != -1)
    {
      gtk_clist_remove ((GtkCList *) clist, selectedrow);
      totalrows--;
      changed = 1;
    }
  else
    {
      warning_dialog ("There is no row selected", 0);
    }
}

void
button_add_clicked (GtkWidget * button, gpointer data)
{
  GtkWidget *window;
  GtkWidget *vbox, *hbox;
  GtkWidget *button2, *button3;
  GtkWidget *label1, *label2, *label3, *label4;
  gchar *text1;

  window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
  gtk_container_border_width (GTK_CONTAINER (window), 10);
  gtk_grab_add (window);
  gtk_window_position (GTK_WINDOW (window), GTK_WIN_POS_MOUSE);
  gtk_window_set_title (GTK_WINDOW (window), "Enter new account");
  gtk_signal_connect (GTK_OBJECT (window), "delete_event",
		      GTK_SIGNAL_FUNC (window_close), (gpointer) window);

  vbox = gtk_vbox_new (FALSE, 5);
  gtk_container_add (GTK_CONTAINER (window), vbox);
  gtk_widget_show (vbox);

  entry1 = gtk_entry_new_with_max_length (100);
  entry2 = gtk_entry_new_with_max_length (50);
  entry3 = gtk_entry_new_with_max_length (50);
  entry4 = gtk_entry_new_with_max_length (100);
  if (selectedrow != -1)
    {
      gtk_clist_get_text (GTK_CLIST (clist), selectedrow, 0, &text1);
      gtk_entry_set_text (GTK_ENTRY (entry1), text1);
      gtk_clist_get_text (GTK_CLIST (clist), selectedrow, 1, &text1);
      gtk_entry_set_text (GTK_ENTRY (entry2), text1);
      gtk_clist_get_text (GTK_CLIST (clist), selectedrow, 2, &text1);
      gtk_entry_set_text (GTK_ENTRY (entry3), text1);
      gtk_clist_get_text (GTK_CLIST (clist), selectedrow, 3, &text1);
      gtk_entry_set_text (GTK_ENTRY (entry4), text1);
    }

  label1 = gtk_label_new ("Server:");
  gtk_box_pack_start (GTK_BOX (vbox), label1, FALSE, FALSE, 0);
  gtk_widget_show (label1);
  gtk_box_pack_start (GTK_BOX (vbox), entry1, FALSE, FALSE, 0);
  gtk_widget_show (entry1);
  label2 = gtk_label_new ("Username:");
  gtk_box_pack_start (GTK_BOX (vbox), label2, FALSE, FALSE, 0);
  gtk_widget_show (label2);
  gtk_box_pack_start (GTK_BOX (vbox), entry2, FALSE, FALSE, 0);
  gtk_widget_show (entry2);
  label3 = gtk_label_new ("Password:");
  gtk_box_pack_start (GTK_BOX (vbox), label3, FALSE, FALSE, 0);
  gtk_widget_show (label3);
  gtk_box_pack_start (GTK_BOX (vbox), entry3, FALSE, FALSE, 0);
  gtk_widget_show (entry3);
  label4 = gtk_label_new ("Comment:");
  gtk_box_pack_start (GTK_BOX (vbox), label4, FALSE, FALSE, 0);
  gtk_widget_show (label4);
  gtk_box_pack_start (GTK_BOX (vbox), entry4, FALSE, FALSE, 0);
  gtk_widget_show (entry4);

  hbox = gtk_hbox_new (TRUE, 0);
  gtk_container_add (GTK_CONTAINER (vbox), hbox);
  gtk_widget_show (hbox);

  button2 = gtk_button_new_with_label ("Cancel");
  button3 = gtk_button_new_with_label ("Add");

  gtk_signal_connect (GTK_OBJECT (button2), "clicked",
		      GTK_SIGNAL_FUNC (window_close),
		      window);
  gtk_signal_connect (GTK_OBJECT (button3), "clicked",
		      GTK_SIGNAL_FUNC (new_set_add),
		      (gpointer) window);
  gtk_box_pack_start (GTK_BOX (hbox), button2, TRUE, TRUE, 0);
  gtk_box_pack_start (GTK_BOX (hbox), button3, TRUE, TRUE, 0);

  gtk_widget_show (button2);
  gtk_widget_show (button3);
  gtk_widget_show (window);
}


void
button_change_clicked (GtkWidget * button, gpointer data)
{

  GtkWidget *vbox, *hbox, *button2, *button3, *window;
  GtkWidget *label1, *label2, *label3, *label4;
  gchar *text1;

  if (selectedrow == -1)
    {
      warning_dialog ("There is no row selected", 0);
      return;
    }

  window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
  gtk_container_border_width (GTK_CONTAINER (window), 10);
  gtk_grab_add (window);
  gtk_window_position (GTK_WINDOW (window), GTK_WIN_POS_MOUSE);
  gtk_window_set_title (GTK_WINDOW (window), "Change account");
  gtk_signal_connect_object (GTK_OBJECT (window), "delete_event",
			     GTK_SIGNAL_FUNC (window_close), (gpointer) window);
  vbox = gtk_vbox_new (FALSE, 5);
  gtk_container_add (GTK_CONTAINER (window), vbox);
  gtk_widget_show (vbox);

  gtk_clist_get_text (GTK_CLIST (clist), selectedrow, 0, &text1);
  entry1 = gtk_entry_new_with_max_length (100);
  gtk_entry_set_text (GTK_ENTRY (entry1), text1);
  gtk_clist_get_text (GTK_CLIST (clist), selectedrow, 1, &text1);
  entry2 = gtk_entry_new_with_max_length (50);
  gtk_entry_set_text (GTK_ENTRY (entry2), text1);
  gtk_clist_get_text (GTK_CLIST (clist), selectedrow, 2, &text1);
  entry3 = gtk_entry_new_with_max_length (50);
  gtk_entry_set_text (GTK_ENTRY (entry3), text1);
  gtk_clist_get_text (GTK_CLIST (clist), selectedrow, 3, &text1);
  entry4 = gtk_entry_new_with_max_length (100);
  gtk_entry_set_text (GTK_ENTRY (entry4), text1);
  label1 = gtk_label_new ("Server:");
  gtk_box_pack_start (GTK_BOX (vbox), label1, FALSE, FALSE, 0);
  gtk_widget_show (label1);
  gtk_box_pack_start (GTK_BOX (vbox), entry1, FALSE, FALSE, 0);
  gtk_widget_show (entry1);
  label2 = gtk_label_new ("Username:");
  gtk_box_pack_start (GTK_BOX (vbox), label2, FALSE, FALSE, 0);
  gtk_widget_show (label2);
  gtk_box_pack_start (GTK_BOX (vbox), entry2, FALSE, FALSE, 0);
  gtk_widget_show (entry2);
  label3 = gtk_label_new ("Password:");
  gtk_box_pack_start (GTK_BOX (vbox), label3, FALSE, FALSE, 0);
  gtk_widget_show (label3);
  gtk_box_pack_start (GTK_BOX (vbox), entry3, FALSE, FALSE, 0);
  gtk_widget_show (entry3);
  label4 = gtk_label_new ("Comment:");
  gtk_box_pack_start (GTK_BOX (vbox), label4, FALSE, FALSE, 0);
  gtk_widget_show (label4);
  gtk_box_pack_start (GTK_BOX (vbox), entry4, FALSE, FALSE, 0);
  gtk_widget_show (entry4);

  hbox = gtk_hbox_new (TRUE, 0);
  gtk_container_add (GTK_CONTAINER (vbox), hbox);
  gtk_widget_show (hbox);

  button2 = gtk_button_new_with_label ("Cancel");
  button3 = gtk_button_new_with_label ("Change");

  gtk_signal_connect_object (GTK_OBJECT (button2), "clicked",
			     GTK_SIGNAL_FUNC (gtk_widget_destroy),
			     (gpointer) window);
  gtk_signal_connect (GTK_OBJECT (button3), "clicked",
		      GTK_SIGNAL_FUNC (change_set),
		      (gpointer) window);
  gtk_box_pack_start (GTK_BOX (hbox), button2, TRUE, TRUE, 0);
  gtk_box_pack_start (GTK_BOX (hbox), button3, TRUE, TRUE, 0);

  gtk_widget_show (button2);
  gtk_widget_show (button3);
  gtk_widget_show (window);
}


void
button_clear_clicked (GtkWidget * button, gpointer data)
{
  GtkWidget *window;
  GtkWidget *label1;
  GtkWidget *button_ok, *button_cancel;
  GtkWidget *vbox, *hbox;

  window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
  gtk_container_border_width (GTK_CONTAINER (window), 10);
  gtk_grab_add (window);
  gtk_window_position (GTK_WINDOW (window), GTK_WIN_POS_MOUSE);
  gtk_window_set_title (GTK_WINDOW (window), "Confirm clear");
  gtk_signal_connect (GTK_OBJECT (window), "delete_event",
		      GTK_SIGNAL_FUNC (window_close), (gpointer) window);

  vbox = gtk_vbox_new (TRUE, 0);
  gtk_container_add (GTK_CONTAINER (window), vbox);
  gtk_widget_show (vbox);

  label1 = gtk_label_new ("Are you sure you want to clear all entries?");
  gtk_box_pack_start (GTK_BOX (vbox), label1, TRUE, TRUE, 0);
  gtk_widget_show (label1);

  hbox = gtk_hbox_new (TRUE, 0);
  gtk_container_add (GTK_CONTAINER (vbox), hbox);
  gtk_widget_show (hbox);

  button_ok = gtk_button_new_with_label ("Clear");
  button_cancel = gtk_button_new_with_label ("Cancel");

  gtk_box_pack_start (GTK_BOX (hbox), button_ok, TRUE, TRUE, 0);
  gtk_widget_show (button_ok);
  gtk_box_pack_start (GTK_BOX (hbox), button_cancel, TRUE, TRUE, 0);
  gtk_widget_show (button_cancel);

  gtk_signal_connect (GTK_OBJECT (button_cancel), "clicked",
		      GTK_SIGNAL_FUNC (window_close),
		      (gpointer) window);
  gtk_signal_connect (GTK_OBJECT (button_ok), "clicked",
		      GTK_SIGNAL_FUNC (really_clear_list),
		      (gpointer) data);
  gtk_signal_connect (GTK_OBJECT (button_ok), "clicked",
		      GTK_SIGNAL_FUNC (window_close),
		      (gpointer) window);

  GTK_WIDGET_SET_FLAGS (button_ok, GTK_CAN_DEFAULT);
  GTK_WIDGET_SET_FLAGS (button_cancel, GTK_CAN_DEFAULT);

  gtk_widget_show (window);
}

void
really_clear_list (GtkWidget * button, gpointer data)
{
  gtk_clist_clear ((GtkCList *) data);
  totalrows = 0;
  changed = 1;
  return;
}

void
change_password (GtkWidget * button, gpointer data)
{
  GtkWidget *window;
  GtkWidget *vbox, *hbox;
  GtkWidget *button2, *button3;
  GtkWidget *label1, *label2, *label3, *label4;

  window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
  gtk_container_border_width (GTK_CONTAINER (window), 10);
  gtk_grab_add (window);
  gtk_window_position (GTK_WINDOW (window), GTK_WIN_POS_MOUSE);
  gtk_window_set_title (GTK_WINDOW (window), "Change password");
  gtk_signal_connect_object (GTK_OBJECT (window), "delete_event",
			     GTK_SIGNAL_FUNC (window_close),
			     (gpointer) window);
  vbox = gtk_vbox_new (FALSE, 0);
  gtk_container_add (GTK_CONTAINER (window), vbox);
  gtk_widget_show (vbox);

  label1 = gtk_label_new ("Change Password");
  label2 = gtk_label_new ("Enter old password");
  entry1 = gtk_entry_new_with_max_length (25);
  gtk_entry_set_visibility (GTK_ENTRY (entry1), 0);
  gtk_entry_set_text (GTK_ENTRY (entry1), "");
  label3 = gtk_label_new ("Enter new password");
  entry2 = gtk_entry_new_with_max_length (25);
  gtk_entry_set_visibility (GTK_ENTRY (entry2), 0);
  gtk_entry_set_text (GTK_ENTRY (entry2), "");
  label4 = gtk_label_new ("Confirm new password");
  entry3 = gtk_entry_new_with_max_length (25);
  gtk_entry_set_text (GTK_ENTRY (entry3), "");
  gtk_entry_set_visibility (GTK_ENTRY (entry3), 0);

  gtk_box_pack_start (GTK_BOX (vbox), label1, TRUE, TRUE, 5);
  gtk_widget_show (label1);
  gtk_box_pack_start (GTK_BOX (vbox), label2, FALSE, FALSE, 0);
  gtk_widget_show (label2);
  gtk_box_pack_start (GTK_BOX (vbox), entry1, FALSE, FALSE, 1);
  gtk_widget_show (entry1);
  gtk_box_pack_start (GTK_BOX (vbox), label3, FALSE, FALSE, 0);
  gtk_widget_show (label3);
  gtk_box_pack_start (GTK_BOX (vbox), entry2, FALSE, FALSE, 1);
  gtk_widget_show (entry2);
  gtk_box_pack_start (GTK_BOX (vbox), label4, FALSE, FALSE, 0);
  gtk_widget_show (label4);
  gtk_box_pack_start (GTK_BOX (vbox), entry3, FALSE, FALSE, 1);
  gtk_widget_show (entry3);

  hbox = gtk_hbox_new (TRUE, 0);
  gtk_container_add (GTK_CONTAINER (vbox), hbox);
  gtk_widget_show (hbox);

  button2 = gtk_button_new_with_label ("Cancel");
  button3 = gtk_button_new_with_label ("Ok");

  gtk_signal_connect (GTK_OBJECT (button2), "clicked",
		      GTK_SIGNAL_FUNC (window_close),
		      (gpointer) window);
  gtk_signal_connect (GTK_OBJECT (button3), "clicked",
		      GTK_SIGNAL_FUNC (check_password),
		      (gpointer) window);
  gtk_box_pack_start (GTK_BOX (hbox), button2, TRUE, TRUE, 0);
  gtk_box_pack_start (GTK_BOX (hbox), button3, TRUE, TRUE, 0);

  GTK_WIDGET_SET_FLAGS (button2, GTK_CAN_DEFAULT);
  GTK_WIDGET_SET_FLAGS (button3, GTK_CAN_DEFAULT);

  gtk_widget_show (button2);
  gtk_widget_show (button3);
#ifdef DEBUG
  g_print ("change_password, window=%p\n", window);
#endif
  gtk_widget_show (window);
}

void
warning_dialog (char *warningtext, char quit)
{

/* if quit ==1 then it will exit gtk_main,
   if quit ==0 then it will just continue */
  GtkWidget *dialog, *label1, *button_ok;
  dialog = gtk_dialog_new ();
  gtk_window_set_title (GTK_WINDOW (dialog), " Warning ");
  gtk_container_set_border_width (GTK_CONTAINER (dialog), 10);
  gtk_window_position (GTK_WINDOW (dialog), GTK_WIN_POS_MOUSE);
  button_ok = gtk_button_new_with_label ("Ok");
  gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->action_area), button_ok,
		      TRUE, TRUE, 2);
  gtk_widget_show (button_ok);
  if (quit)
    {
      gtk_signal_connect (GTK_OBJECT (button_ok), "clicked",
			  GTK_SIGNAL_FUNC (gtk_main_quit),
			  NULL);
    }
  else
    {
      gtk_signal_connect (GTK_OBJECT (button_ok), "clicked",
			  GTK_SIGNAL_FUNC (window_close),
			  (gpointer) dialog);
    }
  label1 = gtk_label_new (warningtext);
  gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->vbox), label1, TRUE,
		      TRUE, 2);
  gtk_widget_show (label1);
  gtk_widget_show (dialog);
}

void
check_password (GtkWidget * button, gpointer window)
{

  char tmppass[30];
  int tmpInt;

  tmpInt = strlen (gtk_entry_get_text (GTK_ENTRY (entry1)));
  tmpInt++;
  strncpy (tmppass, gtk_entry_get_text (GTK_ENTRY (entry1)), tmpInt);

#ifdef DEBUG
  g_print ("check_password, window=%p\n", window);
#endif

  if (strncmp (password, tmppass, tmpInt) != 0)
    {
      window_close (button, window);
      warning_dialog ("Current password incorrect", 0);
      return;
    }

  tmpInt = strlen (gtk_entry_get_text (GTK_ENTRY (entry2)));
  tmpInt++;

  strncpy (tmppass, gtk_entry_get_text (GTK_ENTRY (entry2)), tmpInt);
  if (strncmp (tmppass, gtk_entry_get_text (GTK_ENTRY (entry3)), strlen
      (tmppass)) != 0)
    {
      window_close (button, window);
      warning_dialog ("New passwords are not identical", 0);
      return;
    }
  strncpy (password, gtk_entry_get_text (GTK_ENTRY (entry2)), tmpInt);
  changed = 1;
  window_close (button, window);
}


void
selection_made (GtkWidget * clist, gint row, gint column,
		GdkEventButton * event, gpointer data)
{
  selectedrow = row;
}

void
unselection_made (GtkWidget * clist, gint row, gint column,
		  GdkEventButton * event, gpointer data)
{
  selectedrow = -1;
}

void
click_made (GtkWidget * clist, gint column, gpointer data)
{
  Gpasman_Set *setlist;
  gchar *tmpstr;
  int count;

  setlist = (Gpasman_Set *) g_malloc0 (sizeof (Gpasman_Set) * totalrows);
  for (count = 0; count < totalrows; count++)
    {
      gtk_clist_get_text (GTK_CLIST (clist), count, 0, &tmpstr);
      setlist[count].host = (gchar *) g_malloc0 (sizeof (gchar) * strlen
						 (tmpstr) + 1);
      strcpy (setlist[count].host, tmpstr);
      gtk_clist_get_text (GTK_CLIST (clist), count, 1, &tmpstr);
      setlist[count].username = (gchar *) g_malloc0 (sizeof (gchar) * strlen
						     (tmpstr) + 1);
      strcpy (setlist[count].username, tmpstr);
      gtk_clist_get_text (GTK_CLIST (clist), count, 2, &tmpstr);
      setlist[count].password = (gchar *) g_malloc0 (sizeof (gchar) * strlen
						     (tmpstr) + 1);
      strcpy (setlist[count].password, tmpstr);
      gtk_clist_get_text (GTK_CLIST (clist), count, 3, &tmpstr);
      setlist[count].comment = (gchar *) g_malloc0 (sizeof (gchar) * strlen
						    (tmpstr) + 1);
      strcpy (setlist[count].comment, tmpstr);
    }

  sort_key = column;
  qsort (setlist, totalrows, sizeof (Gpasman_Set), compare_set);
  for (count = 0; count < totalrows; count++)
    {
      gtk_clist_set_text (GTK_CLIST (clist), count, 0, setlist[count].host);
      gtk_clist_set_text (GTK_CLIST (clist), count, 1, setlist[count].username);
      gtk_clist_set_text (GTK_CLIST (clist), count, 2, setlist[count].password);
      gtk_clist_set_text (GTK_CLIST (clist), count, 3, setlist[count].comment);
      free (setlist[count].host);
      free (setlist[count].username);
      free (setlist[count].password);
      free (setlist[count].comment);
    }
  free (setlist);
}


void
save_all (GtkWidget * button, gpointer end)
{

  gchar *text1;
  int MAX_BUFFER_LENGTH = 1024;
  int count1, count2, count3;
  unsigned int j = 0, keylength;
  unsigned char key[128];
  char buffer[MAX_BUFFER_LENGTH];
  unsigned short iv[4];
  unsigned short plaintext[4];
  unsigned short ciphertext[4];
  int bufferIndex;
  FILE *outfile;

  fprintf (stderr, "Saving data\n");

  /* open the outfile (needs better error handling */
#ifdef DEBUG
  g_print ("save_all, started\n");
#endif
/*   strcat(strcpy(naam, getenv("HOME")), "/.gpasman"); */
  outfile = fopen (naam, "wb");
  if (outfile == NULL)
    {
      warning_dialog ("Could not open ~/.gpasman", 0);
      return;
    }

  /* We'll need this to fill the RC2 blocks properly */
  bufferIndex = 0;
  /* make the key ready */
  for (j = 0; password[j] != '\0'; j++)
    {
      key[j] = password[j];
    }
  keylength = j;
  rc2_expandkey (key, keylength, 128);

  /* read the clist into the buffer */
  /* First, we make the IV */
  for (count2 = 0; count2 < 4; count2++)
    {
      iv[count2] = rand ();
      putc ((unsigned char) (iv[count2] >> 8), outfile);
      putc ((unsigned char) (iv[count2] & 0xff), outfile);

    }

  for (count1 = 0; count1 < totalrows; count1++)
    {
      /*  create a random buffer content */
      /* DRIG - why? */
      buffer[0] = '\0';
      srand (key[2]);
      for (count2 = 1; count2 < MAX_BUFFER_LENGTH; count2++)
	{
	  buffer[count2] = 0;	/*rand(); */
	}
      for (count2 = 0; count2 < 4; count2++)
	{
	  gtk_clist_get_text (GTK_CLIST (clist), count1, count2, &text1);
	  if (strlen (text1) == 0)
	    {
	      strncpy (text1, " ", strlen (" "));
	    }
	  strncat (buffer, text1, strlen (text1));
#ifdef DEBUG
	  g_print ("save_all, buffer contains %s\n", buffer);
#endif
	  /* Use 255 as the marker.  \n is too tough to test for */
	  buffer[strlen (buffer)] = 255;
	}

      count2 = 0;

      /* DRIG -
         Okay, I got rid of this confusing mess where the file is being decrypted.
         I'm implementing a different scheme instead.       
         I'm using CBC mode and encrypting the data straight from top down.
         At the bottom, encrypted, I will append an MD5 hash of the file, eventually.
         PKCS 5 padding (explained at the code section
       */

      /* Then, we break the plaintext up into 8 byte chuncks */
      while (count2 < strlen (buffer))
	{
#ifndef WORDS_BIGENDIAN
	  plaintext[bufferIndex] = buffer[count2 + 1] << 8;
	  plaintext[bufferIndex] += buffer[count2] & 0xff;
#endif
#ifdef WORDS_BIGENDIAN
          plaintext[bufferIndex] = buffer[count2] << 8;
          plaintext[bufferIndex] += buffer[count2 + 1] & 0xff;
#endif
	  bufferIndex++;
	  if (bufferIndex == 4)
	    {
	      rc2_encrypt (plaintext);

	      for (count3 = 0; count3 < 4; count3++)
		{
		  ciphertext[count3] = iv[count3] ^ plaintext[count3];

		  /* Now store the ciphertext as the iv */
		  iv[count3] = plaintext[count3];

		  /* reset the buffer index */
		  bufferIndex = 0;
		  putc ((unsigned char) (ciphertext[count3] >> 8), outfile);
		  putc ((unsigned char) (ciphertext[count3] & 0xff), outfile);
		}

	    }
	  // increment a short, not a byte
	  count2 += 2;
	}
    }

  /* Tack on the PKCS 5 padding 
     How it works is we fill up the last n bytes with the value n

     So, if we have, say, 13 bytes, 8 of which are used, we have 5 left 
     over, leaving us 3 short, so we fill it in with 3's.

     If we come out even, we fill it with 8 8s

     um, except that in this instance we are using 4 shorts instead of 8 bytes.
     so, half everything
   */
  for (count1 = bufferIndex; count1 < 4; count1++)
    {
      plaintext[count1] = (4 - bufferIndex);
    }
  rc2_encrypt (plaintext);
  for (count1 = 0; count1 < 4; count1++)
    {
      ciphertext[count1] = iv[count1] ^ plaintext[count1];
      putc ((unsigned char) (ciphertext[count1] >> 8), outfile);
      putc ((unsigned char) (ciphertext[count1] & 0xff), outfile);
    }

  fclose (outfile);
#ifdef DEBUG
      g_print ("save_all, outfile is closed\n");
#endif
  changed = 0;
  if (*(char *) end)
    {
#ifdef DEBUG
      g_print ("save_all, end ==1 do program will exit\n");
#endif
      gtk_main_quit ();
    }
}

int
openfile ()
{
  int count1, count2, count3, count4;
  char buffer[2048], naam[256];
  gchar *entry_text[4];
  unsigned int j = 0, keylength;
  unsigned char key[128];
  FILE *infile;
  unsigned char charbuf[8];
  unsigned int count, size;
  unsigned short iv[4];
  unsigned short plaintext[4];
  unsigned short ciphertext[4];
  int bufferIndex;

  strcat (strcpy (naam, getenv ("HOME")), "/.gpasman");

  /* open the infile (needs better error handling) */
  infile = fopen (naam, "rb");
  if (infile == NULL)
    {
      printf ("Could not open ~/.gpasman");
      warning_dialog ("Could not open ~/.gpasman", 0);
      return -1;
    }

  /* make the key ready */
  for (j = 0; password[j] != '\0'; j++)
    {
      key[j] = password[j];
    }
  keylength = j;
  rc2_expandkey (key, keylength, 128);

  count1 = count2 = count3 = count4 = 0;

/*  not needed, we now know exactly where we are in the buffer
    memset(buffer, '\0', sizeof(buffer));	*/

  count = 0;
  bufferIndex = 0;
  size = read (fileno (infile), (unsigned char *) (charbuf + count), 8);

  if (size < 8)
    return -1;

  for (count = 0; count < 4; count++)
    {
      count2 = count << 1;
      iv[count] = charbuf[count2] << 8;
      iv[count] += charbuf[count2 + 1];
    }

  count = 0;
  size = 0;
  while ((count = read (fileno (infile), (unsigned char *) charbuf, 8)) > 0)
    {
      while (count < 8)
	{
	  count2 = read (fileno (infile), (unsigned char *) (charbuf +
			 count), 8);
	  if (count2 == 0)
	    {
	      printf ("bad EOF\n");
	      return -1;
	    }
	  count += count2;
	}

      size += 8;

      for (count2 = 0; count2 < 8; count2 += 2)
	{
	  count3 = count2 >> 1;
	  ciphertext[count3] = charbuf[count2] << 8;
	  ciphertext[count3] += charbuf[count2 + 1];

	  plaintext[count3] = ciphertext[count3] ^ iv[count3];
	  iv[count3] = plaintext[count3];
	}

      rc2_decrypt (plaintext);
      memcpy ((unsigned char *) (buffer + bufferIndex), plaintext, 8);
      bufferIndex += 8;
      buffer[bufferIndex + 1] = '\0';
    }


  /* Strip off PKCS 5 padding 
     Should check to make sure it's good here
   */
  size -= buffer[size - 1];

  /* This will point to the starting index */
  bufferIndex = 0;
  for (count = 0; count < size; count++)
    {
      if ((unsigned char) (buffer[count]) == 255)
	{
	  if (buffer[bufferIndex] == '\0')
	    {
	      bufferIndex++;
	    }
	  entry_text[count1] = (gchar *) malloc (count - bufferIndex + 1);
	  memcpy (entry_text[count1],
		  (unsigned char *) (buffer + bufferIndex),
		  count - bufferIndex);
	  entry_text[count1][count - bufferIndex] = '\0';
	  count++;
	  bufferIndex = count;
	  count1++;
	  if (count1 == 4)
	    {
	      count1 = 0;
	      gtk_clist_append ((GtkCList *) clist, entry_text);
	      totalrows++;
	    }
	}
    }
  count3 = 0;
  count4 = 0;
  count1 = 0;
  fclose (infile);
#ifdef DEBUG
  g_print ("openfile, ended\n");
#endif
  return 1;
}


int
check_gpasfile (char *passfile)
{
  struct stat naamstat;

  if (stat (passfile, &naamstat) == -1)
    {
      perror ("gpasman, lstat returned -1");
    }

  if (((naamstat.st_mode & S_IAMB) | (S_IRUSR | S_IWUSR)) != (S_IRUSR |
    S_IWUSR))
    {

/*     warning_dialog("~/.gpasman permissions are bad"); */
      fprintf (stderr,
	       "~/.gpasman perms are bad, they are: %ld, should be -rw------\n",
	       (naamstat.st_mode & (S_IREAD | S_IWRITE)));
      return (1);
    }

  if (!S_ISREG (naamstat.st_mode))
    {
      lstat (passfile, &naamstat);
      if (S_ISLNK (naamstat.st_mode))
	{
/*       warning_dialog("~/.gpasman is a symbolic link"); */
	  fprintf (stderr, "~/.gpasman is a symlink\n");

	  return (1);
	}
    }

  return (0);

}

int
file_exists (char *tfile)
{
  struct stat naamstat;

  if ((stat (tfile, &naamstat) == -1) && (errno == ENOENT))
    {
      return (1);
    }
  else
    {
      return (0);
    }
}

int
create_gpasfile ()
{
  mode_t oldumask;
  int ret;

  oldumask = umask (0);

  ret = creat (naam, (S_IRUSR | S_IWUSR));
  umask (oldumask);
  return (ret);
}

int
create_swpfile ()
{
  mode_t oldumask;
  int ret;
  FILE *infile;

  oldumask = umask (0);
  infile = fopen (swp_file, "w");
  if (infile == NULL)
    {
      warning_dialog ("Could not open ~/.gpasman.swp", 0);
      return -1;
    }
  ret = 0;

  fprintf (infile, "%ld\n", getpid ());
  fclose (infile);

  umask (oldumask);
  return (ret);
}

int
remove_swpfile ()
{
  return (unlink (swp_file));
}

int
is_running ()
{
/* this function should check if the PID in the swp file is really a running gpasman */
  int ret;

  if (file_exists(swp_file) ==0 ) {
     ret = 1;
  }
  else 
  {
     ret = 0;
  }
  return (ret);
}


int
compare_set (const void *x, const void *y)
{
  Gpasman_Set *set1 = (Gpasman_Set *) x;
  Gpasman_Set *set2 = (Gpasman_Set *) y;

  switch (sort_key)
    {
    case 0:			/* server */
      return strcmp (set1->host, set2->host);
      break;
    case 1:			/* username */
      return strcmp (set1->username, set2->username);
      break;
    case 2:			/* password */
      return strcmp (set1->password, set2->password);
      break;
    case 3:			/* comment */
      return strcmp (set1->comment, set2->comment);
      break;
    default:			/* wrong column */
      warning_dialog ("Wrong column", 0);
    }				/* switch sort_key */
  return (-1);
}
